
extern crate tobj;
extern crate winit;

use std::f32::consts::PI;
use std::rc::Rc;
use std::sync::Arc;
#[cfg(not(target_arch = "wasm32"))]
use std::time::{Duration, Instant};

use bytemuck::__core::ops::Range;
use cgmath::{Decomposed, Deg, InnerSpace, Quaternion, Rotation3, SquareMatrix};
use futures::task::LocalSpawn;
use legion::*;
use wgpu::{BindGroup, Buffer, TextureView};
use wgpu_subscriber;
use winit::{
    event::{self, WindowEvent},
    event_loop::{ControlFlow, EventLoop},
};
use winit::event::DeviceEvent::MouseMotion;
use winit::platform::unix::x11::ffi::Time;
use rapier3d::math;
use crate::render::Renderer;
use rapier3d::pipeline::PhysicsPipeline;
use rapier3d::dynamics::{IntegrationParameters, RigidBodySet, JointSet, RigidBodyBuilder};
use rapier3d::geometry::{BroadPhase, NarrowPhase, ColliderSet, ColliderBuilder};
use rapier3d::na::{Isometry, Isometry3, Vector, Vector3};

mod geometry;
mod light;
mod render;

/*

Collision detection
https://nphysics.org/rigid_body_simulations_with_contacts/

Obj file format
http://paulbourke.net/dataformats/obj/

tobj obj loader
https://docs.rs/tobj/2.0.3/tobj/index.html

mesh generator lib, might be useful
https://docs.rs/genmesh/0.6.2/genmesh/

legion ECS
https://github.com/amethyst/legion


mvp:

ECS
    animation
    render 3d
    input/io
    collision / physics
    entities & behaviours

 */

#[cfg_attr(rustfmt, rustfmt_skip)]
#[allow(unused)]
pub const OPENGL_TO_WGPU_MATRIX: cgmath::Matrix4<f32> = cgmath::Matrix4::new(
    1.0, 0.0, 0.0, 0.0,
    0.0, 1.0, 0.0, 0.0,
    0.0, 0.0, 0.5, 0.0,
    0.0, 0.0, 0.5, 1.0,
);

// a component is any type that is 'static, sized, send and sync
#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Position {
    x: f32,
    y: f32,
    z: f32,
    mx: cgmath::Matrix4<f32>,
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Color {
    r: f32,
    g: f32,
    b: f32,
    a: f32,
}

#[derive(Clone, Copy, Debug, PartialEq)]
pub struct Velocity {
    dx: f32,
    dy: f32,
    rs: f32,
}

#[derive(Clone, Default, PartialEq, Eq, Hash, Copy, Debug)]
pub struct RangeCopy<Idx> {
    pub start: Idx,
    pub end: Idx,
}

#[derive(Clone, Debug)]
pub struct DirectionalLight {
    color: wgpu::Color,
    fov: f32,
    depth: RangeCopy<f32>,
    target_view: Arc<TextureView>,
}

#[derive(Clone, Debug)]
pub struct Mesh {
    index_buffer: Arc<Buffer>,
    index_count: usize,
    vertex_buffer: Arc<Buffer>,
    uniform_buffer: Arc<Buffer>,
    bind_group: Arc<BindGroup>,
}

//log::info!("");

fn main() {
    let mut world = World::default();

    let (mut pool, spawner) = {
        let local_pool = futures::executor::LocalPool::new();
        let spawner = local_pool.spawner();
        (local_pool, spawner)
    };

    // Schedule for the render systen
    let mut render_schedule = Schedule::builder()
        .add_system(render::render_test_system())
        .build();

    // TODO schedule for the update system and others

    let event_loop = EventLoop::new();
    let mut builder = winit::window::WindowBuilder::new();
    builder = builder.with_title("MVGE");

    // I don't know what they are doing here
    #[cfg(windows_OFF)] // TODO
    {
        use winit::platform::windows::WindowBuilderExtWindows;
        builder = builder.with_no_redirection_bitmap(true);
    }

    let window = builder.build(&event_loop).unwrap();

    let mut last_update_inst = Instant::now();

    // Load up the renderer (and the resources)
    let mut renderer = {
        let mut renderer = render::Renderer::init(&window);
        entity_loading(&mut world, &mut renderer);
        renderer
    };

    let mut resources = Resources::default();
    resources.insert(renderer);

    event_loop.run(move |event, _, control_flow| {
        // Artificially slows the loop rate to 10 millis
        // This is called after redraw events cleared
        *control_flow = ControlFlow::WaitUntil(Instant::now() + Duration::from_millis(10));

        match event {
            event::Event::MainEventsCleared => {
                // ask for a redraw every 20 millis
                if last_update_inst.elapsed() > Duration::from_millis(20) {
                    window.request_redraw();
                    last_update_inst = Instant::now();
                }
                physics();
                pool.run_until_stalled();
            }
            event::Event::DeviceEvent {
                event: MouseMotion { delta },
                ..
            } => {
                resources
                    .get_mut::<Renderer>()
                    .unwrap()
                    .cam_look_delta((delta.0, delta.1));

                //swap_chain = device.create_swap_chain(&surface, &sc_desc);
            }
            // Resizing will queue a request_redraw
            event::Event::WindowEvent {
                event: WindowEvent::Resized(size),
                ..
            } => {
                log::info!("Resizing to {:?}", size);
                let width = size.width;
                let height = size.height;

                resources
                    .get_mut::<Renderer>()
                    .unwrap()
                    .resize(width, height);

                //swap_chain = device.create_swap_chain(&surface, &sc_desc);
            }
            event::Event::WindowEvent { event, .. } => match event {
                WindowEvent::KeyboardInput {
                    input:
                        event::KeyboardInput {
                            virtual_keycode: Some(event::VirtualKeyCode::Escape),
                            state: event::ElementState::Pressed,
                            ..
                        },
                    ..
                }
                | WindowEvent::CloseRequested => {
                    *control_flow = ControlFlow::Exit;
                }
                _ => {
                    //renderer.update(event);
                }
            },
            event::Event::RedrawRequested(_) => {
                // Call the render system
                render_schedule.execute(&mut world, &mut resources);
            }
            _ => {}
        }
    });
}

pub fn physics() {

/*    // let mut mechanical_world = DefaultMechanicalWorld::new(Vector3::new(0.0, -9.81, 0.0));
    // let mut geometrical_world = DefaultGeometricalWorld::new();
    //
    // let mut bodies = DefaultBodySet::new();
    // let mut colliders = DefaultColliderSet::new();
    // let mut joint_constraints = DefaultJointConstraintSet::new();
    // let mut force_generators = DefaultForceGeneratorSet::new();
    //
    // // Run the simulation.
    // mechanical_world.step(
    //     &mut geometrical_world,
    //     &mut bodies,
    //     &mut colliders,
    //     &mut joint_constraints,
    //     &mut force_generators
    // );
    //
    // let ground_handle = bodies.insert(Ground::new());
    // let ground_shape = ShapeHandle::new(Cuboid::new(Vector3::new(3.0, 0.2, 3.0)
    // ));
    //
    // let main_floor = ColliderDesc::new(ground_shape)
    //     .translation(Vector3::y() * -0.2)
    //     .build(BodyPartHandle(ground_handle, 0));
    // colliders.insert(main_floor);
    //
    // // Build the rigid body.
    // let rb = RigidBodyDesc::new()
    //     .translation(Vector3::new(1.0, 1.0, 1.0))
    //     .build();
    // let rb_handle = bodies.insert(rb);
    //
    // // Build the collider.
    // let cuboid = ShapeHandle::new(Cuboid::new(Vector3::new(1.0, 1.0, 1.0)));
    // let co = ColliderDesc::new(cuboid.clone())
    //     .density(1.0)
    //     .build(BodyPartHandle(rb_handle, 0));
    // colliders.insert(co);
    //
    // //testbed.set_body_color(rb_handle, color);
    //
    // let rigid_body =
    //     RigidBodyDesc::new()
    //         // The rigid body position. Will override `.translation(...)` and `.rotation(...)`.
    //         .position(Isometry3::new(
    //             Vector3::new(1.0, 2.0, 3.0),
    //             Vector3::y() * PI,
    //         ))
    //         .gravity_enabled(true)
    //         .status(BodyStatus::Dynamic)
    //         .velocity(VelocityN::linear(1.0, 2.0, 3.0))
    //         .linear_damping(10.0)
    //         .angular_damping(5.0)
    //         .max_linear_velocity(10.0)
    //         .max_angular_velocity(1.7)
    //         .mass(1.2)
    //         // Arbitrary user-defined data associated to the rigid body to be built.
    //         .user_data(10)
    //         .build();
    //
    // let parent_handle = bodies.insert(rigid_body);
    //
    // let shape = ShapeHandle::new(Ball::new(1.5));
    // let collider = ColliderDesc::new(shape)
    //     .density(1.0)
    //     .translation(Vector3::y() * 5.0)
    //     .build(BodyPartHandle(parent_handle, 0));
    // let collider_handle = colliders.insert(collider);*/

    // Here the gravity is -9.81 along the y axis.
    let mut pipeline = PhysicsPipeline::new();
    let gravity = rapier3d::math::Vector::new(0.0, -9.81, 0.0);
    let integration_parameters = IntegrationParameters::default();
    let mut broad_phase = BroadPhase::new();
    let mut narrow_phase = NarrowPhase::new();
    let mut bodies = RigidBodySet::new();
    let mut colliders = ColliderSet::new();
    let mut joints = JointSet::new();
    // We ignore contact events for now.
    let event_handler = ();

    let mut dynamic_ball_body = RigidBodyBuilder::new_dynamic().build();
    let mut dynamic_ball_body_handle = bodies.insert(dynamic_ball_body);

    let mut static_floor_body = RigidBodyBuilder::new_static().position(Isometry3::new(
                Vector3::new(1.0, 2.0, 3.0),
                Vector::y() * PI,
            )).build();
    let mut static_floor_body_handle = bodies.insert(static_floor_body);


    let ball_collider = ColliderBuilder::ball(0.5).build();
    let floor_collider = ColliderBuilder::cuboid(0.5, 0.2, 0.1).build();

    let ball_collider_handle = colliders.insert(ball_collider, dynamic_ball_body_handle, &mut bodies);
    let floor_collider_handle = colliders.insert(floor_collider, dynamic_ball_body_handle, &mut bodies);

    // Run the simulation in the game loop.
    loop {
        pipeline.step(
            &gravity,
            &integration_parameters,
            &mut broad_phase,
            &mut narrow_phase,
            &mut bodies,
            &mut colliders,
            &mut joints,
            None,
            None,
            &event_handler
        );
    }
}

pub fn entity_loading(world: &mut World, renderer: &mut Renderer) {
    let untitled_mesh = renderer.load_mesh_to_buffer("./resources/monkey.obj");

    // This could be used for relationships between entities...???
    let light_entity: Entity = world.push((
        cgmath::Point3 {
            x: 7.0 as f32,
            y: -5.0 as f32,
            z: 10.0 as f32,
        },
        renderer.create_light(),
    ));

    let light_entity: Entity = world.push((
        cgmath::Point3 {
            x: -5.0 as f32,
            y: 7.0 as f32,
            z: 10.0 as f32,
        },
        renderer.create_light(),
    ));

    let offset = cgmath::vec3(2.0, 2.0, 2.0);
    let transform = Decomposed {
        disp: offset.clone(),
        rot: Quaternion::from_axis_angle(offset.normalize(), Deg(50.0)),
        scale: 1.0,
    };

    let mesh_entity: Entity = world.push((
        Position {
            x: 17.0,
            y: 15.0,
            z: 10.0,
            mx: cgmath::Matrix4::from(transform),
        },
        untitled_mesh,
        Color {
            r: 1.0,
            g: 0.5,
            b: 0.5,
            a: 1.0,
        },
    ));

    let plane_mesh = renderer.create_plane(7.0);

    let plane_entity: Entity = world.push((
        Position {
            x: 0.0,
            y: 0.0,
            z: 0.0,
            mx: cgmath::Matrix4::identity(),
        },
        plane_mesh,
        Color {
            r: 1.0,
            g: 0.5,
            b: 0.5,
            a: 1.0,
        },
    ));
}
